@@ -13,18 +13,30 @@ import json
 import logging
 
 
+logger = logging.getLogger('inventree')
+
+
 class InvenTreeAPI(object):
-    """ Basic class for performing Inventree API requests.
-
-    GET - Fetch data from the server
-
     """
+    Basic class for performing Inventree API requests.
+    """
+
+    MIN_SUPPORTED_API_VERSION = 6
+
+    @staticmethod
+    def getMinApiVersion():
+        """
+        Return the minimum supported API version
+        """
+
+        return InvenTreeAPI.MIN_SUPPORTED_API_VERSION
 
     def __init__(self, base_url, **kwargs):
         """ Initialize class with initial parameters
 
         Args:
-            base_url - Base API URL
+            base_url - Base URL for the InvenTree server, including port (if required)
+                       e.g. "http://inventree.server.com:8000"
             
         kwargs:
             username - Login username
@@ -34,19 +46,18 @@ class InvenTreeAPI(object):
             verbose - Print extra debug messages (default = False)
         """
 
-        if not base_url.endswith('/'):
-            base_url += '/'
-
-        # Server address *must* end with /api/
-        if not base_url.endswith('/api/'):
-            base_url = os.path.join(base_url, 'api')
+        # Strip out trailing "/api/" (if provided)
+        if base_url.endswith("/api/"):
+            base_url = base_url[:-5]
 
         if not base_url.endswith('/'):
             base_url += '/'
 
         self.base_url = base_url
 
-        logging.info("Connecting to server: " + str(self.base_url))
+        self.api_url = os.path.join(self.base_url, 'api/')
+
+        logger.info("Connecting to server: " + str(self.base_url))
 
         self.username = kwargs.get('username', None)
         self.password = kwargs.get('password', None)
@@ -67,7 +78,7 @@ class InvenTreeAPI(object):
 
     def clean_url(self, url):
 
-        url = os.path.join(self.base_url, url)
+        url = os.path.join(self.api_url, url)
 
         if not url.endswith('/'):
             url += '/'
@@ -83,16 +94,16 @@ class InvenTreeAPI(object):
 
         self.server_details = None
 
-        logging.info("Checking InvenTree server connection...")
+        logger.info("Checking InvenTree server connection...")
 
         try:
-            response = requests.get(self.base_url)
+            response = requests.get(self.api_url)
         except requests.exceptions.ConnectionError:
-            logging.error("Server connection refused - check server address")
+            logger.error("Server connection refused - check server address")
             return False
 
         if not response.status_code == 200:
-            logging.error("Error code from server: {code} - {detail}".format(
+            logger.error("Error code from server: {code} - {detail}".format(
                 code=response.status_code,
                 detail=response.text
             ))
@@ -102,16 +113,26 @@ class InvenTreeAPI(object):
         # Record server details
         self.server_details = json.loads(response.text)
 
-        logging.info("InvenTree server details: " + str(response.text))
+        logger.info("InvenTree server details: " + str(response.text))
 
         # The details provided by the server should include some specific data:
         server_name = str(self.server_details.get('server', ''))
 
         if not server_name.lower() == 'inventree':
-            logging.warning("Server returned strange response (expected 'InvenTree', found '{name}')".format(
+            logger.warning("Server returned strange response (expected 'InvenTree', found '{name}')".format(
                 name=server_name
             ))
 
+        api_version = self.server_details.get('apiVersion', '1')
+
+        try:
+            api_version = int(api_version)
+        except ValueError:
+            raise ValueError(f"Server returned invalid API version: '{api_version}'")
+
+        if api_version < InvenTreeAPI.getMinApiVersion():
+            raise ValueError(f"Server API version ({api_version}) is older than minimum supported API version ({InvenTreeAPI.getMinApiVersion()})")
+
         return True
 
     def requestToken(self):
@@ -120,31 +141,31 @@ class InvenTreeAPI(object):
         if not self.username or not self.password:
             raise AttributeError('Supply username and password to request token')
 
-        logging.info("Requesting auth token from server...")
+        logger.info("Requesting auth token from server...")
 
         # Request an auth token from the server
-        token_url = os.path.join(self.base_url, 'user/token/')
+        token_url = os.path.join(self.api_url, 'user/token/')
         
         reply = requests.get(token_url, auth=self.auth)
 
         data = json.loads(reply.text)
 
         if not reply.status_code == 200:
-            logging.error("Error requesting token: {code} - {detail}".format(
+            logger.error("Error requesting token: {code} - {detail}".format(
                 code=reply.status_code,
                 detail=str(reply.text)
             ))
             return None
 
         if 'token' not in data.keys():
-            logging.error("Token not returned by server: {detail}".format(
+            logger.error("Token not returned by server: {detail}".format(
                 detail=str(reply.text)
             ))
             return None
 
         self.token = json.loads(reply.text)['token']
 
-        logging.info("Authentication token: " + self.token)
+        logger.info("Authentication token: " + self.token)
 
         return self.token
 
@@ -155,7 +176,7 @@ class InvenTreeAPI(object):
         if url.startswith('/'):
             url = url[1:]
 
-        api_url = os.path.join(self.base_url, url)
+        api_url = os.path.join(self.api_url, url)
 
         if not api_url.endswith('/'):
             api_url += '/'
@@ -166,6 +187,8 @@ class InvenTreeAPI(object):
 
         json = kwargs.get('json', {})
 
+        files = kwargs.get('files', {})
+
         headers = kwargs.get('headers', {})
 
         search_term = kwargs.get('search', None)
@@ -177,11 +200,13 @@ class InvenTreeAPI(object):
             'GET': requests.get,
             'POST': requests.post,
             'PUT': requests.put,
+            'PATCH': requests.patch,
             'DELETE': requests.delete,
+            'OPTIONS': requests.options,
         }
 
         if method.upper() not in methods.keys():
-            logging.error("Unknown request method '{m}'".format(m=method))
+            logger.error("Unknown request method '{m}'".format(m=method))
             return None
 
         method = method.upper()
@@ -192,12 +217,13 @@ class InvenTreeAPI(object):
         else:
             auth = self.auth
 
-        logging.debug("Sending Request:")
-        logging.debug(" - URL:", method, api_url)
-        logging.debug(" - auth:", auth)
-        logging.debug(" - params:", params)
-        logging.debug(" - headers:", headers)
-        logging.debug(" - json:", json)
+        logger.debug("Sending Request:")
+        logger.debug(" - URL:", method, api_url)
+        logger.debug(" - auth:", auth)
+        logger.debug(" - params:", params)
+        logger.debug(" - headers:", headers)
+        logger.debug(" - json:", json)
+        logger.debug(" - files:", files)
 
         try:
             response = methods[method](
@@ -205,23 +231,24 @@ class InvenTreeAPI(object):
                 auth=auth,
                 params=params,
                 headers=headers,
-                json=json
+                json=json,
+                files=files
             )
 
         except requests.exceptions.ConnectionError:
-            logging.error("Connection refused - '{url}'".format(url=api_url))
+            logger.error("Connection refused - '{url}'".format(url=api_url))
             return None
 
         if response is None:
-            logging.error("Null response - {method} '{url}'".format(method=method, url=api_url))
+            logger.error("Null response - {method} '{url}'".format(method=method, url=api_url))
             return None
 
-        logging.info("Request: {method} {url} - {response}".format(method=method, url=api_url, response=response.status_code))
+        logger.info("Request: {method} {url} - {response}".format(method=method, url=api_url, response=response.status_code))
 
         # Detect invalid response codes
         # Anything 300+ is 'bad'
         if response.status_code >= 300:
-            logging.warning("Bad response ({code}) - {method} '{url}' - {detail}".format(
+            logger.warning("Bad response ({code}) - {method} '{url}' - {detail}".format(
                 code=response.status_code, method=method, url=api_url,
                 detail=str(response.text)
             ))
@@ -233,7 +260,7 @@ class InvenTreeAPI(object):
         ctype = response.headers.get('content-type')
 
         if not ctype == 'application/json':
-            logging.error("'Response content-type is not JSON - '{url}' - '{f}'".format(url=api_url, f=ctype))
+            logger.error("'Response content-type is not JSON - '{url}' - '{f}'".format(url=api_url, f=ctype))
             return None
 
         return response
@@ -243,14 +270,19 @@ class InvenTreeAPI(object):
 
         """
 
-        headers = {'content-type': 'application/json'}
+        headers = kwargs.get('headers', {})
 
         response = self.request(url, method='delete', headers=headers, **kwargs)
 
         if response is None:
-            return False
+            return None
 
-        logging.debug(response.status_code, response.text)
+        if response.status_code not in [204]:
+            logger.error(f"DELETE request failed at '{url}' - {response.status_code}")
+
+        logger.debug(response.status_code, response.text)
+
+        return response
 
     def post(self, url, data, files=None, **kwargs):
         """ Perform a POST request. Used to create a new record in the database.
@@ -263,7 +295,7 @@ class InvenTreeAPI(object):
 
         url = self.clean_url(url)
 
-        headers = {}
+        headers = kwargs.get('headers', {})
 
         if self.use_token_auth and self.token:
             headers['AUTHORIZATION'] = 'Token {t}'.format(t=self.token)
@@ -277,61 +309,96 @@ class InvenTreeAPI(object):
             return None
 
         if response.status_code not in [200, 201]:
-            logging.error("POST request failed at '{url}' - {status}".format(url=url, status=response.status_code))
-            logging.error(response.text)
+            logger.error("POST request failed at '{url}' - {status}".format(url=url, status=response.status_code))
+            logger.error(response.text)
+            return None
         
         try:
             data = json.loads(response.text)
         except json.decoder.JSONDecodeError:
-            logging.error("Error decoding JSON response - '{url}'".format(url=url))
+            logger.error("Error decoding JSON response - '{url}'".format(url=url))
             return None
 
         return data
 
-    def put(self, url, data, **kwargs):
-        """ Perform a PUT request. Used to update existing records in the database.
+    def patch(self, url, data, files=None, **kwargs):
+        """
+        Perform a PATCH request.
+
+        Args:
+            url - API endpoint URL
+            data - JSON data
+            files - optional FILES struct
+        """
+
+        headers = kwargs.get('headers', {})
+
+        params = {
+            'format': 'json',
+        }
+
+        response = self.request(
+            url,
+            json=data,
+            method='patch',
+            headers=headers,
+            params=params,
+            files=files,
+            **kwargs
+        )
+
+        if response is None:
+            logger.error(f"PATCH returned null response at '{url}'")
+            return None
+
+        if response.status_code not in [200, 201]:
+            logger.error(f"PATCH request failed at '{url}' - {response.status_code}")
+            return None
+
+        try:
+            data = json.loads(response.text)
+        except json.decoder.JSONDecodeError:
+            logger.error(f"Error decoding JSON response - '{url}'")
+            return None
+
+        return data
+
+    def put(self, url, data, files=None, **kwargs):
+        """
+        Perform a PUT request. Used to update existing records in the database.
 
         Args:
             url - API endpoint URL
             data - JSON data to PUT
         """
 
-        headers = {'content-type': 'application/json'}
+        headers = kwargs.get('headers', {})
 
         params = {
             'format': 'json',
         }
 
-        response = self.request(url, json=data, method='put', headers=headers, params=params, **kwargs)
+        response = self.request(
+            url,
+            json=data,
+            method='put',
+            headers=headers,
+            params=params,
+            files=files,
+            **kwargs
+        )
 
         if response is None:
             return None
-        
+
         if response.status_code not in [200, 201]:
-            logging.error("PUT request failed at '{url}' - {status}".format(url=url, status=response.status_code))
+            logger.error(f"PUT request failed at '{url}' - {response.status_code}")
             return None
 
         try:
             data = json.loads(response.text)
         except json.decoder.JSONDecodeError:
-            logging.error("Error decoding JSON response - '{url}'".format(url=url))
-            return None
-
-        return data
-
-    def put_image(self, url, data, files):
-        headers = {}
-
-        if self.use_token_auth and self.token:
-            headers['AUTHORIZATION'] = 'Token {t}'.format(t=self.token)
-            auth = None
-        else:
-            auth = self.auth
-
-        url = self.clean_url(url)
-        response = requests.put(url, data=data, files=files, headers=headers, auth=auth)
-
-        if response.status_code not in [200, 201]:
+            logger.error(f"Error decoding JSON response - '{url}'")
             return None
 
         return data
@@ -355,7 +422,59 @@ class InvenTreeAPI(object):
         try:
             data = json.loads(response.text)
         except json.decoder.JSONDecodeError:
-            logging.error("Error decoding JSON response - '{url}'".format(url=url))
+            logger.error("Error decoding JSON response - '{url}'".format(url=url))
             return None
 
         return data
+
+    def downloadFile(self, url, destination):
+        """
+        Download a file from the InvenTree server.
+
+        - If the "destination" is a directory, use the filename of the remote URL
+        """
+
+        # Check that the provided URL is "absolute"
+        if not url.startswith(self.base_url):
+
+            if url.startswith('/'):
+                url = url[1:]
+
+            url = os.path.join(self.base_url, url)
+
+        if os.path.exists(destination) and os.path.isdir(destination):
+
+            destination = os.path.join(
+                destination,
+                os.path.basename(url)
+            )
+
+        destination = os.path.abspath(destination)
+
+        headers = {
+            'AUTHORIZATION': f"Token {self.token}"
+        }
+
+        with requests.get(url, stream=True, headers=headers) as request:
+
+            if not request.status_code == 200:
+                logger.error(
+                    f"Error downloading file '{url}': Server returned status {request.status_code}"
+                )
+                return False
+
+            headers = request.headers
+
+            if 'text/html' in headers['Content-Type']:
+                logger.error(
+                    f"Error downloading file '{url}': Server return invalid response (text/html)"
+                )
+                return False
+
+            with open(destination, 'wb') as f:
+
+                for chunk in request.iter_content(chunk_size=16 * 1024):
+                    f.write(chunk)
+
+        logger.info(f"Downloaded '{url}' to '{destination}'")
+        return True
